// -*- C++ -*-
#ifndef HELICITY_AbstractFFTVertex_H
#define HELICITY_AbstractFFTVertex_H
//
// This is the declaration of the AbstractFFTVertex class.
//

#include "VertexBase.h"
#include "ThePEG/Helicity/WaveFunction/SpinorWaveFunction.h"
#include "ThePEG/Helicity/WaveFunction/SpinorBarWaveFunction.h"
#include "ThePEG/Helicity/WaveFunction/TensorWaveFunction.h"
#include "AbstractFFTVertex.fh"

namespace ThePEG {
namespace Helicity {

/**
 * The AbstractFFTVertex class is the base class for all fermion-fermion-tensor
 * interactions in ThePEG
 */
class AbstractFFTVertex: public VertexBase {

public:
  /**
   * Members to calculate the helicity amplitude expressions for vertices
   * and off-shell particles.
   */
  //@{
  /**
   * Evalulate the vertex.
   * @param q2 The scale \f$q^2\f$ for the coupling at the vertex.
   * @param sp1   The wavefunction for the ferimon.
   * @param sbar2 The wavefunction for the antifermion.
   * @param ten3  The wavefunction for the tensor.
   */
  virtual Complex evaluate(Energy2 q2,const SpinorWaveFunction & sp1,
			   const SpinorBarWaveFunction & sbar2, 
			   const TensorWaveFunction & ten3) = 0;
  
  /**
   * Evaluate the off-shell tensor coming from the vertex.
   * @param q2 The scale \f$q^2\f$ for the coupling at the vertex.
   * @param iopt Option of the shape of the Breit-Wigner for the off-shell tensor.
   * @param out The ParticleData pointer for the off-shell tensor.
   * @param sp1   The wavefunction for the ferimon.
   * @param sbar2 The wavefunction for the antifermion.
   * @param mass The mass of the off-shell particle if not taken from the ParticleData
   * object
   * @param width The width of the off-shell particle if not taken from the ParticleData
   * object
   */
  virtual TensorWaveFunction evaluate(Energy2 q2, int iopt,tcPDPtr out,
				      const SpinorWaveFunction & sp1,
				      const SpinorBarWaveFunction & sbar2,
				      Energy mass=-GeV, Energy width=-GeV) = 0;

  /**
   * Evaluate the off-shell spinor coming from the vertex.
   * @param q2 The scale \f$q^2\f$ for the coupling at the vertex.
   * @param iopt Option of the shape of the Breit-Wigner for the off-shell spinor.
   * @param out The ParticleData pointer for the off-shell spinor.
   * @param sp1   The wavefunction for the ferimon.
   * @param ten3  The wavefunction for the tensor.
   * @param mass The mass of the off-shell particle if not taken from the ParticleData
   * object
   * @param width The width of the off-shell particle if not taken from the ParticleData
   * object
   * @param drep The Dirac matrix representation
   */
  virtual SpinorWaveFunction evaluate(Energy2 q2,int iopt,tcPDPtr out,
				      const SpinorWaveFunction & sp1,
				      const TensorWaveFunction & ten3,
				      Energy mass=-GeV, Energy width=-GeV,
				      DiracRep drep=defaultDRep) = 0;

  /**
   * Evaluate the off-shell barred spinor coming from the vertex.
   * @param q2 The scale \f$q^2\f$ for the coupling at the vertex.
   * @param iopt Option of the shape of the Breit-Wigner for the off-shell barred spinor.
   * @param out The ParticleData pointer for the off-shell barred spinor.
   * @param sbar2 The wavefunction for the antifermion.
   * @param ten3  The wavefunction for the tensor.
   * @param mass The mass of the off-shell particle if not taken from the ParticleData
   * object
   * @param width The width of the off-shell particle if not taken from the ParticleData
   * object
   * @param drep The Dirac matrix representation
   */
  virtual SpinorBarWaveFunction evaluate(Energy2 q2,int iopt, tcPDPtr out,
					 const SpinorBarWaveFunction & sbar2,
					 const TensorWaveFunction& ten3,
					 Energy mass=-GeV, Energy width=-GeV,
					 DiracRep drep=defaultDRep) = 0;
  //@}

public:

  /**
   * The standard Init function used to initialize the interfaces.
   * Called exactly once for each class by the class description system
   * before the main function starts or
   * when this class is dynamically loaded.
   */
  static void Init();

private:

  /**
   * The static object used to initialize the description of this class.
   * Indicates that this is an abstract class without persistent data.
   */
  static AbstractNoPIOClassDescription<AbstractFFTVertex> initAbstractFFTVertex;

  /**
   * The assignment operator is private and must never be called.
   * In fact, it should not even be implemented.
   */
  AbstractFFTVertex & operator=(const AbstractFFTVertex &);

};

}
}

#include "ThePEG/Utilities/ClassTraits.h"

namespace ThePEG {

/** @cond TRAITSPECIALIZATIONS */

/** This template specialization informs ThePEG about the
 *  base classes of AbstractFFTVertex. */
template <>
struct BaseClassTrait<Helicity::AbstractFFTVertex,1> {
  /** Typedef of the first base class of AbstractFFTVertex. */
  typedef Helicity::VertexBase NthBase;
};

/** This template specialization informs ThePEG about the name of
 *  the AbstractFFTVertex class and the shared object where it is defined. */
template <>
struct ClassTraits<Helicity::AbstractFFTVertex>
  : public ClassTraitsBase<Helicity::AbstractFFTVertex> {
  /** Return a platform-independent class name */
  static string className() { return "Helicity::AbstractFFTVertex"; }
};

/** @endcond */

}

#endif /* HELICITY_AbstractFFTVertex_H */
